
import logging

import xlwings as xw
from xlcalculator import ModelCompiler
from xlcalculator import Model
from xlcalculator import Evaluator
import numpy as np
import pandas as pd

logging.basicConfig(level=logging.INFO)

workbook = xw.books.active

if 'FlyingKoala.conf' in [ sheet.name for sheet in workbook.sheets ]:
    auto_load_value = workbook.sheets['FlyingKoala.conf'].range('B3').value
else:
    auto_load_value = False
    logging.error("Would be great to see a worksheet FlyingKoala.conf")

excel_file_name = workbook.fullname
excel_model = None # xlcalculator model
ignore_sheets = "FlyingKoala.conf\\_FlyingKoala.conf\\xlwings.conf\\_xlwings.conf"
koala_models = {}

@xw.sub
def generate_model_graph(model, refresh = False):
    """The function that extracts a graph of a given model from the Spreadsheet"""
    global koala_models
    global excel_model

    if isinstance(model, str):
        if refresh == False and model in koala_models.keys():
            return 'Model %s is already cached, set refresh True if you want it to refresh it' % model.name.name
    else:
        if refresh == False and model.name.name in koala_models.keys():
            return 'Model %s is already cached, set refresh True if you want it to refresh it' % model.name.name

    # logging.debug(parser.prettyprint())
    inputs = [model.name.name]
    inputs.extend(excel_model.formulae[model.name.name].terms)
    koala_models[str(model.name.name)] = ModelCompiler.extract(excel_model, focus=inputs)

    logging.info("Successfully loaded model {}".format(model.name))
    return 'Cached Model %s' % model.name


@xw.sub
def reload_koala(file_name, ignore_sheets= None, bootstrap_equations= None):
    """Loads the Excel workbook into a Python compatible object"""
    global excel_model

    ignore_sheets = ignore_sheets.split('\\')

    logging.info("Loading workbook")

    if file_name is '':
        logging.debug("file_name is not set in Excel Ribbon using {}".format(excel_file_name))
        file_name = excel_file_name

    excel_compiler = ModelCompiler()
    excel_model = excel_compiler.read_and_parse_archive(file_name, ignore_sheets=ignore_sheets)

    logging.info("Workbook '{}' has been loaded.".format(file_name))
    logging.info("Ignored worksheets {}".format(ignore_sheets))


if auto_load_value == True:
    ignore_sheets = "{}\\{}".format(workbook.sheets['FlyingKoala.conf'].range('B2').value, ignore_sheets)
    reload_koala(excel_file_name, ignore_sheets=ignore_sheets)


@xw.func
def reset_koala_model_cache():
    global koala_models

    koala_models = {}


@xw.func
def get_model_cache_count():
    global koala_models

    return len(koala_models.keys())


@xw.func
def get_named_range_count():

    wb = xw.books.active
    return len(wb.names)


@xw.func
def get_cached_koala_model_names():
    global koala_models

    names_of_cached_models = ""
    for model_name in koala_models.keys():
        names_of_cached_models += "\r\n%s" % (model_name)

    return names_of_cached_models


@xw.func
def get_named_ranges():

    returnable = ""
    wb = xw.books.active
    for name in wb.names:
        returnable += "\r\n%s" % (name.name)

    return returnable


@xw.func
@xw.arg('model_name', doc='Name, as a string, of the model which might be cached.')
def is_koala_model_cached(model_name):
    global koala_models

    return model_name in koala_models.keys()


def load_model(model_name):
    """Preparing model name from either a string or an xlwings Range and load it into cache."""
    global koala_models
    global excel_model

    if excel_model is None:
        reload_koala(excel_file_name, ignore_sheets=ignore_sheets)

    # figure out if we have a named range or a text name of the model
    extracted_model_name = None
    if model_name is not None:
        if model_name.name is None:
            extracted_model_name = model_name.value
        else:
            extracted_model_name = model_name.name.name
    else:
        raise Exception('The range you tried to use does not exist in the workbook, if named range exists check spelling.')

    # ensure that model is cached
    if extracted_model_name not in koala_models.keys():
        model = None
        wb = xw.books.active
        for name in wb.names:
            if extracted_model_name == name.name:
                model = xw.Range(extracted_model_name)
                generate_model_graph(model)

        if model is None:
            raise Exception('Model "%s" has not been loaded into cache, if named range exists check spelling.' % extracted_model_name)

    return extracted_model_name


@xw.func
@xw.arg('model_name', doc='Name, as a string, of the model which might be cached.')
def unload_koala_model_from_cache(model_name):
    global koala_models

    del(koala_models[model_name])


def evaluate_koala_model_row(model_name, input_data, model, no_calc_when_zero=[]):
    """The function which sets the input values in the model and evaluates the Excel equation using koala"""
    global koala_models
    model = koala_models[str(model_name)]
    for key in input_data.keys():
        if key in no_calc_when_zero and input_data[key] == 0:
            return

        model.set_cell_value(key, input_data[key])

    evaluator = Evaluator(model)

    return evaluator.evaluate(model_name)


@xw.sub
def evaluate_koala_model(model_name, terms, no_calc_when_zero=[]):
    """The function that sets up the evaluation of the koala equation"""
    global koala_models
    global excel_compiler

    if model_name not in koala_models.keys():
        return 'Model %s has not been loaded into cache.' % model_name

    def eval(row, model):
        return evaluate_koala_model_row(model_name, row.to_dict(), model, no_calc_when_zero)

    current_model = koala_models[model_name]

    results = terms.apply(eval, axis= 1, model= current_model)
    return results




from datetime import timedelta

import xlwings as xw
import numpy as np
import pandas as pd

from flyingkoala import flyingkoala as fk

@xw.func
@xw.arg('times', np.array, doc='This is the range of times')
@xw.arg('inputs', np.array, doc='This is the value you want to average from.')
@xw.arg('window', doc='The number of elements which will be averaged.')
@xw.ret(index=False, header=False, expand='down')
def TIMESERIESWINDOWAVERAGE(times, inputs, window=5):
    """Performs a look-ahead moving average of size window on a time series for values in inputs"""
    def include_average(time, value, window_size):
        ts = pd.to_datetime(str(time))
        mymod = np.mod(int(ts.strftime('%M')), window_size)
        if mymod == np.int32(0):
            return value

    window_size = int(window)

    timeseries = pd.DataFrame({'time':times, 'value':inputs})
    upside_down = timeseries.iloc[::-1]
    upside_down['average'] = upside_down['value'].rolling(window_size).mean()
    timeseries = upside_down.iloc[::-1]

    timeseries['returnable'] = np.vectorize(include_average)(timeseries['time'], timeseries['average'], window_size)

    return timeseries['returnable']


@xw.func
@xw.arg('times', np.array, doc='This is the range of times')
@xw.arg('inputs', np.array, doc='This is the value you want to average from.')
@xw.arg('window', doc='The elements will be kept on this index.')
@xw.ret(expand='down')
def KEEPRECORDS(times, inputs, window=5):
    """Keeps records at an offset determined by window"""

    returnable = []
    ascending = True
    timeseries = pd.DataFrame({'time':times, 'value':inputs})

    delta = timedelta(minutes = window)

    if timeseries.index.is_monotonic_increasing:
        ascending = False

    goal_time = None
    for index, row in timeseries.iterrows():
        if goal_time == None:
            returnable.append([row['value']])
            if ascending:
                goal_time = row['time'] + delta
            else:
                goal_time = row['time'] - delta
        else:

            if goal_time - row['time'] == timedelta(milliseconds = 0):
                returnable.append([row['value']])
                if ascending:
                    goal_time = goal_time + delta
                else:
                    goal_time = goal_time - delta
            else:
                returnable.append([None])

    return returnable


@xw.func
@xw.arg('times', np.array, doc='This is the range of times')
@xw.arg('inputs', np.array, doc='This is the value you want to average from.')
@xw.arg('window', doc='The number of elements which will be averaged.')
@xw.arg('operation', doc='The operation by which the resample will occur.')
@xw.ret(index=False, header=False, expand='down')
def RESAMPLEMINS(times, inputs, window=5, operation='mean'):
    """Performs a look-ahead re-sample of size window with stated operation on a time series for values in inputs and does not return the time index"""
    window_size = int(window)

    timeseries = pd.DataFrame({'time':times, 'value':inputs})
    timeseries.set_index('time', inplace=True)
    upside_down = timeseries.iloc[::-1]
    if operation == 'mean':
        thing = upside_down.resample('{0}Min'.format(int(window))).mean()
    elif operation == 'sum':
        thing = upside_down.resample('{0}Min'.format(int(window))).sum()

    return thing['value']


@xw.func
@xw.arg('times', np.array, doc='This is the range of times')
@xw.arg('inputs', np.array, doc='This is the value you want to average from.')
@xw.arg('window', doc='The number of elements which will be averaged.')
@xw.arg('operation', doc='The operation by which the resample will occur.')
@xw.ret(index=True, header=False, expand='down')
def RESAMPLEMINSWITHINDEX(times, inputs, window=5, operation='mean'):
    """Performs a look-ahead re-sample average of size window on a time series for values in inputs and returns with the time index"""
    window_size = int(window)

    timeseries = pd.DataFrame({'time':times, 'value':inputs})
    timeseries.set_index('time', inplace=True)
    # upside_down = timeseries.iloc[::-1]
    thing = upside_down.resample('{0}Min'.format(int(window))).mean()

    return thing['value']


@xw.func
@xw.arg('keys', np.array, doc='The reference time.')
@xw.arg('below', doc='Integer number of hours leading up to the key time.')
@xw.arg('above', doc='Integer number of hours beyond the key time.')
@xw.arg('periods', np.array, doc='The times which will be determined within the period')
@xw.ret(index=False, header=False, expand='down')
def TIMEISBETWEEN(keys, below, above, periods):
    """Decides if a time is between certain range of a given time"""

    def include_period(key, below_delta, above_delta, period):
        below_date = pd.to_datetime(str(key)) - below_delta
        above_date = pd.to_datetime(str(key)) + above_delta
        if pd.to_datetime(str(period)) >= below_date and pd.to_datetime(str(period)) < above_date:
            return True

    below_delta = timedelta(hours=below)
    above_delta = timedelta(hours=above)

    time_between = pd.DataFrame({'keys': keys, 'periods': periods})

    time_between['returnable'] = np.vectorize(include_period)(time_between['keys'], below_delta, above_delta, time_between['periods'])

    return time_between['returnable']



# accounting.py
import xlwings as xw
from xlwings import constants

from harvest import Harvest
from harvest.reports import Reports
from harvest.dataclasses import *

import .utils

personal_access_token = PersonalAccessToken('ACCOUNT ID', 'PERSONAL ACCESS TOKEN')
reports = Reports("https://api.harvestapp.com/api/v2", personal_access_token)

def get_harvest_invoices():
    wb = xw.Book.caller()

    sheets = xw.sheets

    invoice_sheet = None
    for sheet in sheets:
        if sheet.name == 'Harvest Invoices':
            invoice_sheet = sheet

    if invoice_sheet is not None:
        invoice_sheet.delete()

    new_invoices = sheets.add(name='Harvest Invoices', after='Sheet1')

    wb.app.screen_updating = False
    wb.app.calculation = 'manual'

    new_invoices.range('A1').value = 'client.id'
    new_invoices.range('B1').value = 'id'
    new_invoices.range('C1').value = 'number'
    new_invoices.range('D1').value = 'purchase_order'
    new_invoices.range('E1').value = 'tax'
    new_invoices.range('F1').value = 'tax_amount'
    new_invoices.range('G1').value = 'tax2'
    new_invoices.range('H1').value = 'tax2_amount'
    new_invoices.range('I1').value = 'discount'
    new_invoices.range('J1').value = 'discount_amount'
    new_invoices.range('K1').value = 'period_start'
    new_invoices.range('L1').value = 'period_end'
    new_invoices.range('M1').value = 'paid_date'
    new_invoices.range('N1').value = 'closed_at'
    new_invoices.range('O1').value = 'paid_at'
    new_invoices.range('P1').value = 'estimate'
    new_invoices.range('Q1').value = 'retainer'
    new_invoices.range('R1').value = 'sent_at'
    new_invoices.range('S1').value = 'notes'
    new_invoices.range('T1').value = 'client_key'
    new_invoices.range('U1').value = 'amount'
    new_invoices.range('V1').value = 'due_amount'
    new_invoices.range('W1').value = 'subject'
    new_invoices.range('X1').value = 'state'
    new_invoices.range('Y1').value = 'issue_date'
    new_invoices.range('Z1').value = 'due_date'
    new_invoices.range('AA1').value = 'payment_term'
    new_invoices.range('AB1').value = 'created_at'
    new_invoices.range('AC1').value = 'updated_at'
    new_invoices.range('AD1').value = 'currency'
    new_invoices.range('AE1').value = 'creator.id'
    new_invoices.range('AF1').value = 'creator.name'
    new_invoices.range('AG1').value = 'client.name'

    new_invoices.range('AH1').value = 'line_item.id'
    new_invoices.range('AI1').value = 'line_item.kind'
    new_invoices.range('AJ1').value = 'line_item.description'
    new_invoices.range('AK1').value = 'line_item.quantity'
    new_invoices.range('AL1').value = 'line_item.unit_price'
    new_invoices.range('AM1').value = 'line_item.amount'
    new_invoices.range('AN1').value = 'line_item.taxed'
    new_invoices.range('AO1').value = 'line_item.taxed2'

    new_invoices.range('AP1').value = 'line_item.project.id'
    new_invoices.range('AQ1').value = 'line_item.project.name'

    invoices = reports.invoices()

    row_counter = 2
    for invoice in invoices.invoices:
        for line_item in invoice.line_items:
            new_invoices.range('A%s' % row_counter).value = invoice.client.id
            new_invoices.range('B%s' % row_counter).value = invoice.id
            new_invoices.range('C%s' % row_counter).value = invoice.number
            new_invoices.range('D%s' % row_counter).value = invoice.purchase_order
            new_invoices.range('E%s' % row_counter).value = invoice.tax
            new_invoices.range('F%s' % row_counter).value = invoice.tax_amount
            new_invoices.range('G%s' % row_counter).value = invoice.tax2
            new_invoices.range('H%s' % row_counter).value = invoice.tax2_amount
            new_invoices.range('I%s' % row_counter).value = invoice.discount
            new_invoices.range('J%s' % row_counter).value = invoice.discount_amount
            new_invoices.range('K%s' % row_counter).value = invoice.period_start
            new_invoices.range('L%s' % row_counter).value = invoice.period_end
            new_invoices.range('M%s' % row_counter).value = invoice.paid_date
            new_invoices.range('N%s' % row_counter).value = invoice.closed_at
            new_invoices.range('O%s' % row_counter).value = invoice.paid_at
            new_invoices.range('P%s' % row_counter).value = invoice.estimate
            new_invoices.range('Q%s' % row_counter).value = invoice.retainer
            new_invoices.range('R%s' % row_counter).value = invoice.sent_at
            new_invoices.range('S%s' % row_counter).value = invoice.notes
            new_invoices.range('T%s' % row_counter).value = invoice.client_key
            new_invoices.range('U%s' % row_counter).value = invoice.amount
            new_invoices.range('V%s' % row_counter).value = invoice.due_amount
            new_invoices.range('W%s' % row_counter).value = invoice.subject
            new_invoices.range('X%s' % row_counter).value = invoice.state
            new_invoices.range('Y%s' % row_counter).value = invoice.issue_date
            new_invoices.range('Z%s' % row_counter).value = invoice.due_date
            new_invoices.range('AA%s' % row_counter).value = invoice.payment_term
            new_invoices.range('AB%s' % row_counter).value = invoice.created_at
            new_invoices.range('AC%s' % row_counter).value = invoice.updated_at
            new_invoices.range('AD%s' % row_counter).value = invoice.currency
            new_invoices.range('AE%s' % row_counter).value = invoice.creator.id
            new_invoices.range('AF%s' % row_counter).value = invoice.creator.name
            new_invoices.range('AG%s' % row_counter).value = invoice.client.name

            new_invoices.range('AH%s' % row_counter).value = line_item.id
            new_invoices.range('AI%s' % row_counter).value = line_item.kind
            new_invoices.range('AJ%s' % row_counter).value = line_item.description
            new_invoices.range('AK%s' % row_counter).value = line_item.quantity
            new_invoices.range('AL%s' % row_counter).value = line_item.unit_price
            new_invoices.range('AM%s' % row_counter).value = line_item.amount
            new_invoices.range('AN%s' % row_counter).value = line_item.taxed
            new_invoices.range('AO%s' % row_counter).value = line_item.taxed2

            if line_item.project is not None:
                new_invoices.range('AP%s' % row_counter).value = line_item.project.id
                new_invoices.range('AQ%s' % row_counter).value = line_item.project.name

            row_counter += 1

    wb.app.screen_updating = True
    wb.app.calculation = 'automatic'


def get_harvest_time_entries():
    wb = xw.Book.caller()

    sheets = xw.sheets

    time_entries_sheet = None
    for sheet in sheets:
        if sheet.name == 'Harvest Time Entries':
            time_entries_sheet = sheet

    if time_entries_sheet is not None:
        time_entries_sheet.delete()

    new_time_entries = sheets.add(name='Harvest Time Entries', after='Sheet1')

    wb.app.screen_updating = False
    wb.app.calculation = 'manual'

    new_time_entries.range('A1').value = 'notes' # OPTIONAL
    new_time_entries.range('B1').value = 'locked_reason' # OPTIONAL
    new_time_entries.range('C1').value = 'timer_started_at' # OPTIONAL
    new_time_entries.range('D1').value = 'started_time' # OPTIONAL
    new_time_entries.range('E1').value = 'ended_time' # OPTIONAL
    new_time_entries.range('F1').value = 'invoice' #COMPUND OBJECT # OPTIONAL
    new_time_entries.range('G1').value = 'external_reference' # OPTIONAL
    new_time_entries.range('H1').value = 'external_reference' # OPTIONAL
    new_time_entries.range('I1').value = 'billable_rate' # OPTIONAL
    new_time_entries.range('J1').value = 'id'
    new_time_entries.range('K1').value = 'spent_date'
    new_time_entries.range('L1').value = 'user.id'
    new_time_entries.range('M1').value = 'user.name'
    new_time_entries.range('N1').value = 'client.id'
    new_time_entries.range('O1').value = 'client.name'
    new_time_entries.range('P1').value = 'project.id'
    new_time_entries.range('Q1').value = 'project.name'
    new_time_entries.range('R1').value = 'task.id'
    new_time_entries.range('S1').value = 'task.name'
    new_time_entries.range('T1').value = 'user_assignment.id'
    new_time_entries.range('U1').value = 'user_assignment.is_project_manager'
    new_time_entries.range('V1').value = 'user_assignment.is_active'
    new_time_entries.range('W1').value = 'user_assignment.budget'
    new_time_entries.range('X1').value = 'user_assignment.created_at'
    new_time_entries.range('Y1').value = 'user_assignment.updated_at'
    new_time_entries.range('Z1').value = 'user_assignment.hourly_rate'
    new_time_entries.range('AA1').value = 'task_assignment.id'
    new_time_entries.range('AB1').value = 'task_assignment.billable'
    new_time_entries.range('AC1').value = 'task_assignment.is_active'
    new_time_entries.range('AD1').value = 'task_assignment.created_at'
    new_time_entries.range('AE1').value = 'task_assignment.updated_at'
    new_time_entries.range('AF1').value = 'task_assignment.jourly_rate'
    new_time_entries.range('AG1').value = 'task_assignment.budget'
    new_time_entries.range('AH1').value = 'hours'
    new_time_entries.range('AI1').value = 'created_at'
    new_time_entries.range('AJ1').value = 'is_locked'
    new_time_entries.range('AK1').value = 'is_closed'
    new_time_entries.range('AL1').value = 'is_billed'
    new_time_entries.range('AM1').value = 'is_running'
    new_time_entries.range('AN1').value = 'billable'
    new_time_entries.range('AO1').value = 'budgeted'
    new_time_entries.range('AP1').value = 'cost_rate'

    time_entries = reports.time_entries()

    row_counter = 2
    for time_entry in time_entries.time_entries:
        new_time_entries.range('A%s' % row_counter).value = time_entry.notes # OPTIONAL
        new_time_entries.range('B%s' % row_counter).value = time_entry.locked_reason # OPTIONAL
        new_time_entries.range('C%s' % row_counter).value = time_entry.timer_started_at # OPTIONAL
        new_time_entries.range('D%s' % row_counter).value = time_entry.started_time # OPTIONAL
        new_time_entries.range('E%s' % row_counter).value = time_entry.ended_time # OPTIONAL
        new_time_entries.range('F%s' % row_counter).value = time_entry.invoice #COMPUND OBJECT # OPTIONAL
        new_time_entries.range('G%s' % row_counter).value = time_entry.external_reference # OPTIONAL
        new_time_entries.range('H%s' % row_counter).value = time_entry.external_reference # OPTIONAL
        new_time_entries.range('I%s' % row_counter).value = time_entry.billable_rate # OPTIONAL
        new_time_entries.range('J%s' % row_counter).value = time_entry.id
        new_time_entries.range('K%s' % row_counter).value = time_entry.spent_date
        new_time_entries.range('L%s' % row_counter).value = time_entry.user.id
        new_time_entries.range('M%s' % row_counter).value = time_entry.user.name
        new_time_entries.range('N%s' % row_counter).value = time_entry.client.id
        new_time_entries.range('O%s' % row_counter).value = time_entry.client.name
        new_time_entries.range('P%s' % row_counter).value = time_entry.project.id
        new_time_entries.range('Q%s' % row_counter).value = time_entry.project.name
        new_time_entries.range('R%s' % row_counter).value = time_entry.task.id
        new_time_entries.range('S%s' % row_counter).value = time_entry.task.name
        new_time_entries.range('T%s' % row_counter).value = time_entry.user_assignment.id
        new_time_entries.range('U%s' % row_counter).value = time_entry.user_assignment.is_project_manager
        new_time_entries.range('V%s' % row_counter).value = time_entry.user_assignment.is_active
        new_time_entries.range('W%s' % row_counter).value = time_entry.user_assignment.budget
        new_time_entries.range('X%s' % row_counter).value = time_entry.user_assignment.created_at
        new_time_entries.range('Y%s' % row_counter).value = time_entry.user_assignment.updated_at
        new_time_entries.range('Z%s' % row_counter).value = time_entry.user_assignment.hourly_rate
        new_time_entries.range('AA%s' % row_counter).value = time_entry.task_assignment.id
        new_time_entries.range('AB%s' % row_counter).value = time_entry.task_assignment.billable
        new_time_entries.range('AC%s' % row_counter).value = time_entry.task_assignment.is_active
        new_time_entries.range('AD%s' % row_counter).value = time_entry.task_assignment.created_at
        new_time_entries.range('AE%s' % row_counter).value = time_entry.task_assignment.updated_at
        new_time_entries.range('AF%s' % row_counter).value = time_entry.task_assignment.hourly_rate
        new_time_entries.range('AG%s' % row_counter).value = time_entry.task_assignment.budget
        new_time_entries.range('AH%s' % row_counter).value = time_entry.hours
        new_time_entries.range('AI%s' % row_counter).value = time_entry.created_at
        new_time_entries.range('AJ%s' % row_counter).value = time_entry.is_locked
        new_time_entries.range('AK%s' % row_counter).value = time_entry.is_closed
        new_time_entries.range('AL%s' % row_counter).value = time_entry.is_billed
        new_time_entries.range('AM%s' % row_counter).value = time_entry.is_running
        new_time_entries.range('AN%s' % row_counter).value = time_entry.billable
        new_time_entries.range('AO%s' % row_counter).value = time_entry.budgeted
        new_time_entries.range('AP%s' % row_counter).value = time_entry.cost_rate

        row_counter += 1

    wb.app.screen_updating = True
    wb.app.calculation = 'automatic'



from datetime import datetime

import xlwings as xw
import numpy as np
import pandas as pd

from flyingkoala import flyingkoala as fk

@xw.func
@xw.arg('file_name', 'raw', doc='Full path file name for the .axf file.')
@xw.ret(expand='table', ndim=2, index=None)
def load_observations_axf(file_name):
    """Function to extract the data table from a Bureau of Meteorology observations .axf"""

    with open(file_name, 'r') as file:
        data = False
        header = True
        df = None
        header_cells = None
        returnable = []

        for line in file:
            if line == '[$]\n':
                data = False

            if data:
                if header:
                    header_cells = line.split(',')
                    returnable.append(header_cells)
                    header = False
                else:
                    record_cells = line.split(',')
                    record_cells[2] = record_cells[2][1:-1]
                    record_cells[3] = record_cells[3][1:-1]
                    record_cells[5] = datetime.strptime(record_cells[5],'"%Y%m%d%H%M%S"')
                    record_cells[6] = datetime.strptime(record_cells[6],'"%Y%m%d%H%M%S"')
                    returnable.append(record_cells)

            if line == '[data]\n':
                data = True

    return returnable




from datetime import *

import xlwings as xw
import numpy as np
import pandas as pd
import pvlib
from timezonefinder import TimezoneFinder
import pytz

import flyingkoala
from flyingkoala import *

timezone_finder = TimezoneFinder()

@xw.func
@xw.arg('latitude', doc='The latitude of the desired location')
@xw.arg('longitude', doc='The longitude of the desired location')
@xw.arg('times', np.array, doc='This is the range of times')
@xw.ret(index=False, header=False, transpose=True, expand='down')
def SOLARNOON(latitude, longitude, times):
    """Finds solar noon for each day in teh series"""
    # global timezone_finder
    # timezone_name = timezone_finder.timezone_at(lng=longitude, lat=latitude)
    # tz = timezone(timezone_name)

    float_lat = float(latitude)
    float_lon = float(longitude)

    if not isinstance(times[0], datetime):
        times = pd.to_datetime(times)

    if times[0].tzinfo is not None and times[0].tzinfo.utcoffset(times[0]) is not None:
        print("timezone aware")
    else:
        print("timezone naieve")

    solar_noon = pd.DataFrame({'time_local':times})

    day_lookup = []
    def days(time):
        date_only = pd.to_datetime(str(time)).strftime('%Y-%m-%d')
        if date_only not in day_lookup:
            day_lookup.append(date_only)
        return date_only
    solar_noon['date_local'] = np.vectorize(days)(solar_noon['time_local'])

    solar_noons = {}
    for day in day_lookup:
        this_days_series = solar_noon[ solar_noon['date_local'] == day ]
        solar_position = pvlib.solarposition.get_solarposition(this_days_series['time_local'], float_lat, float_lon, method='nrel_numpy')
        solar_noons[day] = solar_position['zenith'].idxmin()
    solar_noon['zenith_local'] = solar_noon['date_local'].map(solar_noons)

    def drop_timezone(zenith_utc):
        return zenith_utc.replace(tzinfo=None)

    solar_noon['zenith_local_sans_tz'] = solar_noon['zenith_local'].apply(drop_timezone)

    return np.vectorize(pd.Timestamp)(solar_noon['zenith_local_sans_tz'].values)



# utils.py
import xlwings as xw
from xlwings import constants
from openpyxl import utils


def delete_columns(sheet_to_delete_from, columns_to_keep, sheet_rename, delete=True):
    if not isinstance(columns_to_keep, list):
        return

    # get the worksheet we are interested in (providing it exists)
    wb = xw.Book.caller()
    sheets = xw.sheets
    working_sheet = None
    for sheet in sheets:
        if sheet.name == sheet_to_delete_from:
            working_sheet = sheet

    if working_sheet is not None:
        columns = working_sheet.used_range.columns.count

        # excel column references are 1 based
        columns += 1
        length = 0
        column_ordinals_to_delete = []

        for col in range(1, columns):
            length += 1
            # are we getting rid of these columns, or keeping them and deleting the others..?
            if delete:
                if working_sheet.cells(1, col).value not in columns_to_keep:
                    column_ordinals_to_delete.append(length)
            else:
                if working_sheet.cells(1, col).value in columns_to_keep:
                    column_ordinals_to_delete.append(length)

        # delete from the right-hand-side first so the ordinals don't change.
        column_ordinals_to_delete.reverse()

        for column in column_ordinals_to_delete:
            letter = utils.cell.get_column_letter(column)
            working_sheet.range('%s:%s' % (letter, letter)).api.Delete(constants.DeleteShiftDirection.xlShiftToLeft)

        working_sheet.name = sheet_rename

def keep_columns(sheet_to_delete_from, columns_to_delete, sheet_rename):
    return delete_columns(sheet_to_delete_from, columns_to_delete, sheet_rename, delete=False)


def employee_rate_and_total(sheet_range = 'Employee Hours', sheet_lookup = 'Employee Rates'):

    # get the worksheet we are interested in (providing it exists)
    wb = xw.Book.caller()
    sheets = xw.sheets
    working_sheet = None
    for sheet in sheets:
        if sheet.name == sheet_range:
            working_sheet = sheet

    if working_sheet is not None:
        Lookup_cols = working_sheet.used_range.columns.count
        Lookup_rows = working_sheet.used_range.rows.count + 1

        rates_col = Lookup_cols + 1                   #calculates the column number for the rates
        total_col = Lookup_cols + 2                    #calculates the row number for the rates
        rows = working_sheet.used_range.rows.count             #calculates the number of rows in "Sheet1"
        code_position = 0
        surname_position = 0
        hours_position = 0

        for col in range(1, Lookup_cols + 1):
            if working_sheet.cells(1, col).value == 'Project Code':
                code_position = col

            if working_sheet.cells(1, col).value == 'Last Name':
                surname_position = col

            if working_sheet.cells(1, col).value == 'Hours':
                hours_position = col

        rates_Letter = utils.cell.get_column_letter(rates_col)
        total_Letter = utils.cell.get_column_letter(total_col)
        code_Letter = utils.cell.get_column_letter(code_position)
        surname_Letter = utils.cell.get_column_letter(surname_position)
        hours_Letter = utils.cell.get_column_letter(hours_position)

        working_sheet.cells(1,rates_col).value = "rates"
        working_sheet.cells(2,rates_col).value = "=VLOOKUP(%s%s, '%s'!$A$1:$B$%s, 2, FALSE)" % (surname_Letter, 2, sheet_lookup, Lookup_rows)
        working_sheet.cells(2,rates_col).api.autofill(working_sheet.range("%s2:%s%s" % (rates_Letter, rates_Letter, rows)).api, constants.AutoFillType.xlFillDefault)

        working_sheet.cells(1,total_col).value = "total"
        working_sheet.cells(2,total_col).value = "=%s2*%s2" %(hours_Letter, rates_Letter)
        working_sheet.cells(2,total_col).api.autofill(working_sheet.range("%s2:%s%s" % (total_Letter, total_Letter, rows)).api, constants.AutoFillType.xlFillDefault)

def make_employee_pivot_table():
    # get the worksheet we are interested in (providing it exists)
    wb = xw.Book.caller()
    sheets = xw.sheets
    working_sheet = None
    for sheet in sheets:
        if sheet.name == 'Employee Hours':
            working_sheet = sheet

    for col in range(1, working_sheet.used_range.columns.count + 1):
        if working_sheet.cells(1, col).value == 'Project Code':
            code_position = col

        if working_sheet.cells(1, col).value == 'Last Name':
            surname_position = col

        if working_sheet.cells(1, col).value == 'total':
            total_position = col

    code_Letter = utils.cell.get_column_letter(code_position)
    surname_Letter = utils.cell.get_column_letter(surname_position)
    total_Letter = utils.cell.get_column_letter(total_position)

    pivot_data = xw.sheets.add(name='Pivot Data', after='Sheet1')

    if working_sheet is not None:
        working_sheet.range("%s:%s" % (code_Letter, code_Letter)).api.Copy(pivot_data.range("A:A").api )
        working_sheet.range("%s:%s" % (surname_Letter, surname_Letter)).api.Copy(pivot_data.range("B:B").api )

        working_sheet.range("%s:%s" % (total_Letter, total_Letter)).api.Copy()
        pivot_data.range("C1").api.PasteSpecial(Paste = constants.PasteType.xlPasteValues)

        val_rows = pivot_data.used_range.rows.count                             #counts the number of rows in the values sheet

        PivotSourceRange = pivot_data.range("A1:C%s" % val_rows)                #selects the source data for the pivot table

        pivot_table = xw.sheets.add(name='Pivot Table', after='Sheet1')

        PivotTableName = 'ReportPivotTable'

        PivotCache = wb.api.PivotCaches().Create(SourceType=constants.PivotTableSourceType.xlDatabase, SourceData=PivotSourceRange.api, Version=constants.PivotTableVersionList.xlPivotTableVersion14)
        PivotTable = PivotCache.CreatePivotTable(TableDestination="'Pivot Table'!R1C1", TableName=PivotTableName, DefaultVersion=constants.PivotTableVersionList.xlPivotTableVersion14)
        PivotTable.PivotFields('Last Name').Orientation = constants.PivotFieldOrientation.xlRowField
        PivotTable.PivotFields('Last Name').Position = 1
        PivotTable.PivotFields('Project Code').Orientation = constants.PivotFieldOrientation.xlRowField
        PivotTable.PivotFields('Project Code').Position = 2
        PivotTable.PivotFields("total").Orientation = constants.PivotFieldOrientation.xlDataField

        piv_rows = pivot_table.used_range.rows.count                             #counts the number of rows in the pivot table

        people = []
        for each in PivotTable.PivotFields("Last Name").PivotItems():     #creates a list containing the names of all the people in the pivot table
            people.append(each.value)

        the_name = None
        for row in range(1, piv_rows):                                   #matches the name to the corresping code
            if (pivot_table.cells(row, 1).value in people):
                the_name = pivot_table.cells(row, 1).value
            else:
                pivot_table.cells(row, 3).value = the_name

        people = ["Row Labels", "Grand Total"] + people                                #adds 'Row Labels' and 'Grand Total' to the start and finish of the people list

        pivot_text = xw.sheets.add(name='Pivot Text', after='Sheet1')
        pivot_text.cells(1, 1).value = 'Project Code'
        pivot_text.cells(1, 2).value = 'total'
        pivot_text.cells(1, 3).value = 'Last Name'

        #copies and pastes the the pivot table to the pivot_text sheet
        pivot_table.range("A1:C%s" % piv_rows).api.Copy()
        pivot_text.range("A2").api.PasteSpecial(Paste = constants.PasteType.xlPasteValues)

        for row in reversed( range(1, piv_rows + 1) ):
            if pivot_text.cells(row, 1).value in people:
                pivot_text.cells(row, 1).api.EntireRow.Delete()

def project_code_and_names(sheet_range = 'Pivot Text', sheet_lookup = 'Project Code and Names'):

    # get the worksheet we are interested in (providing it exists)
    wb = xw.Book.caller()
    sheets = xw.sheets
    working_sheet = None
    for sheet in sheets:
        if sheet.name == sheet_range:
            working_sheet = sheet

    if working_sheet is not None:
        working_sheet.cells(1, 4).value = 'Project Code and Name'

        rows = working_sheet.used_range.rows.count             #calculates the number of rows in "Sheet1"

        working_sheet.cells(2,4).value = "=VLOOKUP(A2, '%s'!$A:$B, 2, FALSE)" % (sheet_lookup)
        working_sheet.cells(2,4).api.autofill(working_sheet.range("$D$2:$D$%s" % rows).api, constants.AutoFillType.xlFillDefault)



from .timeseries import TIMESERIESWINDOWAVERAGE
from .timeseries import KEEPRECORDS
from .timeseries import RESAMPLEMINS
from .timeseries import RESAMPLEMINSWITHINDEX
from .timeseries import TIMEISBETWEEN



"""
FlyingKoala provides the ability to dynamically define Python calculations from Excel formulas enabling you to replace the Excel calculation engine with Python with a genuine minimum of Python code.

The main benefit of replacing existing formulas with FyingKoala is that you get faster calculation, less need to get your hands dirty and can unit test your formulas. It speeds up scenario analysis and model development

Two users;
Those who want to easily add python functions to excel
Those who want to speed up their existing crazy big models

"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="flyingkoala",
    version="0.0.9b0",
    author="Bradley van Ree",
    author_email="flyingkoala@bradbase.net",
    description="Integration of xlwings and xlcalculator with MS Excel plug-in",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['xls',
        'excel',
        'spreadsheet',
        'workbook',
        'vba',
        'macro',
        'data analysis',
        'analysis'
        'reading excel',
        'excel formula',
        'excel formulas',
        'excel equations',
        'excel equation',
        'formula',
        'formulas',
        'equation',
        'equations',
        'pandas',
        'harvest',
        'timeseries',
        'time series',
        'energy',
        'accounting',
        'horticulture',
        'research',
        'visualization',
        'scenario analysis',
        'modelling',
        'model',
        'unit testing',
        'testing',
        'audit'],
    url="https://github.com/bradbase/flyingkoala",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
    ],
    install_requires=[
            'xlwings >= 0.15.8',
            'numpy >= 1.15.0',
            'pandas >= 0.25.0',
            'openpyxl >= 3.0.3',
            'python-harvest-apiv2 >= 1.0.1',
            'networkx >= 2.4',
            'xlcalculator >= 0.0.2b'
        ]
)



name = "flyingkoala"
from .flyingkoala import generate_model_graph
from .flyingkoala import reload_koala
from .flyingkoala import reset_koala_model_cache
from .flyingkoala import get_cached_koala_model_names
from .flyingkoala import unload_koala_model_from_cache
from .flyingkoala import evaluate_koala_model
from .flyingkoala import get_model_cache_count
from .flyingkoala import get_named_range_count
from .flyingkoala import get_named_ranges
from .flyingkoala import load_model



import xlwings as xw
import numpy as np
import pandas as pd

from flyingkoala import flyingkoala as fk


@xw.func
@xw.arg('model_name', doc='Name, as a string, of the model which will be evaluated. The Excel cell name / named range')
@xw.arg('T_min', np.array, doc='Daily minimum temperature')
@xw.arg('T_max', np.array, doc='Daily maximum temperature')
@xw.ret(index=False, header=False)
def DegreeDayByName(model_name, T_min, T_max):
    """Function to assemble a dataframe for calculating Degree Day.

    :param model: Text name (string) of the model which will be evaluated. The Excel cell name / named range
    :param T_min: Daily minimum temperature
    :param T_max: Daily maximum temperature
    """

    if model_name not in fk.koala_models.keys():
        model = None
        wb = xw.books.active
        for name in wb.names:
            if model_name == name.name:
                model = xw.Range(model_name)
                fk.generate_model_graph(model)

        if model is None:
            return 'Model "%s" has not been loaded into cache, if named range exists check spelling.' % model_name

    inputs_for_DegreeDay = pd.DataFrame({'T_min': np.array([T_min]), 'T_max': np.array([T_max])})
    return fk.evaluate_koala_model(model_name, inputs_for_DegreeDay)


@xw.func
@xw.arg('model_name', doc='Name, as a string, of the model which will be evaluated. The Excel cell name / named range')
@xw.arg('T_min', np.array, doc='Daily minimum temperature')
@xw.arg('T_max', np.array, doc='Daily maximum temperature')
@xw.ret(index=False, header=False, expand='down')
def DegreeDayDynamicArrayByName(model_name, T_min, T_max):
    """Function to assemble a dataframe for calculating Degree Day using dynamic arrays.

    :param model: Text name (string) of the model which will be evaluated. The Excel cell name / named range
    :param T_min: Daily minimum temperature
    :param T_max: Daily maximum temperature
    """

    if model_name not in fk.koala_models.keys():
        model = None
        wb = xw.books.active
        for name in wb.names:
            if model_name == name.name:
                model = xw.Range(model_name)
                fk.generate_model_graph(model)

        if model is None:
            return 'Model "%s" has not been loaded into cache, if named range exists check spelling.' % model_name

    inputs_for_DegreeDay = pd.DataFrame({'T_min': T_min, 'T_max': T_max})
    return fk.evaluate_koala_model(model_name, inputs_for_DegreeDay)


@xw.func
@xw.arg('model_name', xw.Range, doc='Name, as a string or named range, of the model which will be evaluated.')
@xw.arg('T_min', np.array, doc='Daily minimum temperature')
@xw.arg('T_max', np.array, doc='Daily maximum temperature')
@xw.ret(index=False, header=False)
def DegreeDay(model_name, T_min, T_max):
    """Function to assemble a dataframe for calculating Degree Day. Takes either a name of a model or the named range itself.

    :param model: Named Range of the model which will be evaluated. The Excel cell name / named range
    :param T_min: Daily minimum temperature
    :param T_max: Daily maximum temperature
    """
    try:
        prepared_model_name = fk.load_model(model_name)

        # format the equation terms, evalueate the equation and return the result.
        inputs_for_DegreeDay = pd.DataFrame({'T_min': np.array([T_min]), 'T_max': np.array([T_max])})
        return fk.evaluate_koala_model(prepared_model_name, inputs_for_DegreeDay)

    except Exception as err:
        return str(err)


@xw.func
@xw.arg('model_name', xw.Range, doc='Name, as a string or named range, of the model which will be evaluated.')
@xw.arg('T_min', np.array, doc='Daily minimum temperature')
@xw.arg('T_max', np.array, doc='Daily maximum temperature')
@xw.ret(index=False, header=False, expand='down')
def DegreeDayDynamicArray(model_name, T_min, T_max):
    """Function to assemble a dataframe for calculating Degree Day using dynamic arrays. Takes either a name of a model or the named range itself.

    :param model: Named Range of the model which will be evaluated. The Excel cell name / named range
    :param T_min: Daily minimum temperature
    :param T_max: Daily maximum temperature
    """
    try:
        prepared_model_name = fk.load_model(model_name)

        # format the equation terms, evalueate the equation and return the result.
        inputs_for_DegreeDay = pd.DataFrame({'T_min': T_min, 'T_max': T_max})
        return fk.evaluate_koala_model(prepared_model_name, inputs_for_DegreeDay)

    except Exception as err:
            return str(err)
